'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; /**
                                                                                                                                                                                                                                                                   *  Copyright (c) Facebook, Inc.
                                                                                                                                                                                                                                                                   *  All rights reserved.
                                                                                                                                                                                                                                                                   *
                                                                                                                                                                                                                                                                   *  This source code is licensed under the license found in the
                                                                                                                                                                                                                                                                   *  LICENSE file in the root directory of this source tree.
                                                                                                                                                                                                                                                                   *
                                                                                                                                                                                                                                                                   *  
                                                                                                                                                                                                                                                                   */

exports.processIPCNotificationMessage = processIPCNotificationMessage;
exports.processIPCRequestMessage = processIPCRequestMessage;
exports.processStreamMessage = processStreamMessage;

var _GraphQLCache = require('./GraphQLCache');

var _getOutline = require('../interfaces/getOutline');

var _GraphQLLanguageService = require('../interfaces/GraphQLLanguageService');

var _GraphQLConfig = require('../config/GraphQLConfig');

var _Range = require('../utils/Range');

// Response message error codes
var ERROR_CODES = {
  PARSE_ERROR: -32700,
  INVALID_REQUEST: -32600,
  METHOD_NOT_FOUND: -32601,
  INVALID_PARAMS: -32602,
  INTERNAL_ERROR: -32603,
  SERVER_ERROR_START: -32099,
  SERVER_ERROR_END: -32000,
  SERVER_NOT_INITIALIZED: -32002,
  UNKNOWN_ERROR_CODE: -32001
};

var REQUEST_IDS_IN_PROGRESS = [];

var graphQLCache = void 0;
var languageService = void 0;
var textDocumentCache = new Map();

function processIPCNotificationMessage(message) {
  return regeneratorRuntime.async(function processIPCNotificationMessage$(_context) {
    while (1) {
      switch (_context.prev = _context.next) {
        case 0:
          throw new Error('Not yet implemented.');

        case 1:
        case 'end':
          return _context.stop();
      }
    }
  }, null, this);
}

function processIPCRequestMessage(message, configDir) {
  var method, response, textDocument, serverCapabilities, uri, text, cachedDocument, diagnostics, contentChanges, _cachedDocument, position, _cachedDocument2, query, _result, requestIDToCancel, index;

  return regeneratorRuntime.async(function processIPCRequestMessage$(_context2) {
    while (1) {
      switch (_context2.prev = _context2.next) {
        case 0:
          method = message.method;
          response = void 0;
          textDocument = void 0;
          _context2.t0 = method;
          _context2.next = _context2.t0 === 'initialize' ? 6 : _context2.t0 === 'textDocument/didOpen' ? 14 : _context2.t0 === 'textDocument/didSave' ? 14 : _context2.t0 === 'textDocument/didChange' ? 26 : _context2.t0 === 'textDocument/didClose' ? 32 : _context2.t0 === 'textDocument/completion' ? 37 : _context2.t0 === '$/cancelRequest' ? 50 : _context2.t0 === 'shutdown' ? 56 : _context2.t0 === 'exit' ? 57 : 59;
          break;

        case 6:
          if (!(!message.params || !message.params.rootPath)) {
            _context2.next = 8;
            break;
          }

          return _context2.abrupt('return');

        case 8:
          _context2.next = 10;
          return regeneratorRuntime.awrap(initialize(configDir ? configDir.trim() : message.params.rootPath));

        case 10:
          serverCapabilities = _context2.sent;


          if (serverCapabilities === null) {
            response = convertToRpcMessage({
              id: '-1',
              error: {
                code: ERROR_CODES.SERVER_NOT_INITIALIZED,
                message: '.graphqlrc not found'
              }
            });
          } else {
            response = convertToRpcMessage({
              id: message.id,
              result: serverCapabilities
            });
          }
          sendMessageIPC(response);

          return _context2.abrupt('break', 59);

        case 14:
          if (!(!message.params || !message.params.textDocument)) {
            _context2.next = 16;
            break;
          }

          return _context2.abrupt('return');

        case 16:
          textDocument = message.params.textDocument;
          uri = textDocument.uri;
          text = textDocument.text;

          if (!text) {
            if (textDocumentCache.has(textDocument.uri)) {
              cachedDocument = textDocumentCache.get(textDocument.uri);

              if (cachedDocument) {
                text = cachedDocument.content.text;
              }
            }
          }

          _context2.next = 22;
          return regeneratorRuntime.awrap(provideDiagnosticsMessage(text, uri));

        case 22:
          diagnostics = _context2.sent;

          response = convertToRpcMessage({
            id: message.id,
            method: 'textDocument/publishDiagnostics',
            params: { uri: uri, diagnostics: diagnostics }
          });
          sendMessageIPC(response);
          return _context2.abrupt('break', 59);

        case 26:
          if (!(!message.params || !message.params.textDocument || !message.params.contentChanges)) {
            _context2.next = 28;
            break;
          }

          return _context2.abrupt('return');

        case 28:

          textDocument = message.params.textDocument;
          contentChanges = message.params.contentChanges;


          if (textDocumentCache.has(textDocument.uri)) {
            _cachedDocument = textDocumentCache.get(textDocument.uri);

            if (_cachedDocument && _cachedDocument.version < textDocument.version) {
              // Current server capabilities specify the full sync of the contents.
              // Therefore always overwrite the entire content.
              // Also, as `contentChanges` is an array and we just want the
              // latest update to the text, grab the last entry from the array.
              textDocumentCache.set(textDocument.uri, {
                version: textDocument.version,
                content: contentChanges[contentChanges.length - 1]
              });
            }
          } else {
            textDocumentCache.set(textDocument.uri, {
              version: textDocument.version,
              content: contentChanges[contentChanges.length - 1]
            });
          }
          return _context2.abrupt('break', 59);

        case 32:
          if (!(!message.params || !message.params.textDocument)) {
            _context2.next = 34;
            break;
          }

          return _context2.abrupt('return');

        case 34:
          textDocument = message.params.textDocument;

          if (textDocumentCache.has(textDocument.uri)) {
            textDocumentCache.delete(textDocument.uri);
          }
          return _context2.abrupt('break', 59);

        case 37:
          if (!(!message.params || !message.params.textDocument || !message.params.position)) {
            _context2.next = 39;
            break;
          }

          return _context2.abrupt('return');

        case 39:
          textDocument = message.params.textDocument;
          position = message.params.position;

          if (!textDocumentCache.has(textDocument.uri)) {
            _context2.next = 49;
            break;
          }

          _cachedDocument2 = textDocumentCache.get(textDocument.uri);

          if (!_cachedDocument2) {
            _context2.next = 49;
            break;
          }

          query = _cachedDocument2.content.text;
          _context2.next = 47;
          return regeneratorRuntime.awrap(languageService.getAutocompleteSuggestions(query, position, textDocument.uri));

        case 47:
          _result = _context2.sent;


          sendMessageIPC(convertToRpcMessage({
            id: message.id,
            result: _result
          }));

        case 49:
          return _context2.abrupt('break', 59);

        case 50:
          if (!(!message.params || !message.params.id)) {
            _context2.next = 52;
            break;
          }

          return _context2.abrupt('return');

        case 52:
          requestIDToCancel = message.params.id;
          index = REQUEST_IDS_IN_PROGRESS.indexOf(requestIDToCancel);

          if (index !== -1) {
            REQUEST_IDS_IN_PROGRESS.splice(index, 1);
            // A cancelled request still needs to send an empty response back
            sendMessageIPC({ id: requestIDToCancel });
          }
          return _context2.abrupt('break', 59);

        case 56:
          return _context2.abrupt('break', 59);

        case 57:
          process.exit(0);
          return _context2.abrupt('break', 59);

        case 59:
        case 'end':
          return _context2.stop();
      }
    }
  }, null, this);
}

function processStreamMessage(message, configDir) {
  var graphQLConfigDir, json, id, method, result, responseMsg, _json$args, query, filePath, position, formatted;

  return regeneratorRuntime.async(function processStreamMessage$(_context3) {
    while (1) {
      switch (_context3.prev = _context3.next) {
        case 0:
          if (!(message.length === 0)) {
            _context3.next = 2;
            break;
          }

          return _context3.abrupt('return');

        case 2:
          if (graphQLCache) {
            _context3.next = 12;
            break;
          }

          _context3.next = 5;
          return regeneratorRuntime.awrap((0, _GraphQLConfig.findGraphQLConfigDir)(configDir ? configDir.trim() : process.cwd()));

        case 5:
          graphQLConfigDir = _context3.sent;

          if (graphQLConfigDir) {
            _context3.next = 9;
            break;
          }

          process.stdout.write(JSON.stringify(convertToRpcMessage({
            id: '-1',
            error: {
              code: ERROR_CODES.SERVER_NOT_INITIALIZED,
              message: '.graphqlrc not found'
            }
          })));
          return _context3.abrupt('return');

        case 9:
          _context3.next = 11;
          return regeneratorRuntime.awrap((0, _GraphQLCache.getGraphQLCache)(graphQLConfigDir));

        case 11:
          graphQLCache = _context3.sent;

        case 12:
          if (!languageService) {
            languageService = new _GraphQLLanguageService.GraphQLLanguageService(graphQLCache);
          }

          json = void 0;
          _context3.prev = 14;

          json = JSON.parse(message);
          _context3.next = 22;
          break;

        case 18:
          _context3.prev = 18;
          _context3.t0 = _context3['catch'](14);

          process.stdout.write(JSON.stringify(convertToRpcMessage({
            id: '-1',
            error: {
              code: ERROR_CODES.PARSE_ERROR,
              message: 'Request contains incorrect JSON format'
            }
          })));
          return _context3.abrupt('return');

        case 22:
          id = json.id;
          method = json.method;
          result = null;
          responseMsg = null;
          _json$args = json.args, query = _json$args.query, filePath = _json$args.filePath, position = _json$args.position;
          _context3.t1 = method;
          _context3.next = _context3.t1 === 'disconnect' ? 30 : _context3.t1 === 'getDiagnostics' ? 32 : _context3.t1 === 'getDefinition' ? 38 : _context3.t1 === 'getAutocompleteSuggestions' ? 44 : _context3.t1 === 'getOutline' ? 51 : 55;
          break;

        case 30:
          process.exit(0);
          return _context3.abrupt('break', 56);

        case 32:
          _context3.next = 34;
          return regeneratorRuntime.awrap(provideDiagnosticsMessage(query, filePath));

        case 34:
          result = _context3.sent;

          responseMsg = convertToRpcMessage({
            type: 'response',
            id: id,
            result: result
          });
          process.stdout.write(JSON.stringify(responseMsg) + '\n');
          return _context3.abrupt('break', 56);

        case 38:
          _context3.next = 40;
          return regeneratorRuntime.awrap(languageService.getDefinition(query, position, filePath));

        case 40:
          result = _context3.sent;

          responseMsg = convertToRpcMessage({
            type: 'response',
            id: id,
            result: result
          });
          process.stdout.write(JSON.stringify(responseMsg) + '\n');
          return _context3.abrupt('break', 56);

        case 44:
          _context3.next = 46;
          return regeneratorRuntime.awrap(languageService.getAutocompleteSuggestions(query, position, filePath));

        case 46:
          result = _context3.sent;
          formatted = result.map(function (res) {
            return {
              text: res.label,
              typeName: res.detail ? String(res.detail) : null,
              description: res.documentation || null
            };
          });

          responseMsg = convertToRpcMessage({
            type: 'response',
            id: id,
            formatted: formatted
          });
          process.stdout.write(JSON.stringify(responseMsg) + '\n');
          return _context3.abrupt('break', 56);

        case 51:
          result = (0, _getOutline.getOutline)(query);
          responseMsg = convertToRpcMessage({
            type: 'response',
            id: id,
            result: result
          });
          process.stdout.write(JSON.stringify(responseMsg) + '\n');
          return _context3.abrupt('break', 56);

        case 55:
          return _context3.abrupt('break', 56);

        case 56:
        case 'end':
          return _context3.stop();
      }
    }
  }, null, this, [[14, 18]]);
}

/**
 * Helper functions to perform requested services from client/server.
 */

function initialize(rootPath) {
  var serverCapabilities, configDir;
  return regeneratorRuntime.async(function initialize$(_context4) {
    while (1) {
      switch (_context4.prev = _context4.next) {
        case 0:
          serverCapabilities = {
            capabilities: {
              completionProvider: { resolveProvider: true },
              textDocumentSync: 1
            }
          };
          _context4.next = 3;
          return regeneratorRuntime.awrap((0, _GraphQLConfig.findGraphQLConfigDir)(rootPath));

        case 3:
          configDir = _context4.sent;

          if (configDir) {
            _context4.next = 6;
            break;
          }

          return _context4.abrupt('return', null);

        case 6:
          _context4.next = 8;
          return regeneratorRuntime.awrap((0, _GraphQLCache.getGraphQLCache)(configDir));

        case 8:
          graphQLCache = _context4.sent;

          languageService = new _GraphQLLanguageService.GraphQLLanguageService(graphQLCache);

          return _context4.abrupt('return', serverCapabilities);

        case 11:
        case 'end':
          return _context4.stop();
      }
    }
  }, null, this);
}

function provideDiagnosticsMessage(query, uri) {
  var results;
  return regeneratorRuntime.async(function provideDiagnosticsMessage$(_context5) {
    while (1) {
      switch (_context5.prev = _context5.next) {
        case 0:
          _context5.next = 2;
          return regeneratorRuntime.awrap(languageService.getDiagnostics(query, uri));

        case 2:
          results = _context5.sent;

          if (results && results.length > 0) {
            (function () {
              var queryLines = query.split('\n');
              var totalLines = queryLines.length;
              var lastLineLength = queryLines[totalLines - 1].length;
              var lastCharacterPosition = new _Range.Position(totalLines, lastLineLength);
              results = results.filter(function (diagnostic) {
                return diagnostic.range.end.lessThanOrEqualTo(lastCharacterPosition);
              });
            })();
          }

          return _context5.abrupt('return', results);

        case 5:
        case 'end':
          return _context5.stop();
      }
    }
  }, null, this);
}

function sendMessageIPC(message) {
  if (process.send !== undefined) {
    process.send(message);
  }
}

/**
 * Composes a language server protocol JSON message.
 */
function convertToRpcMessage(metaMessage) {
  var message = _extends({
    jsonrpc: '2.0',
    protocol: 'graphql_language_service'
  }, metaMessage);

  return message;
}